from unittest import main

from tests.test_case import CommonTestCase
from tests.model import TestObject
from marshmallow_advanced import Schema, fields, ValidationError


class TestToInstance(CommonTestCase):

    @classmethod
    def setUpClass(cls) -> None:
        super(TestToInstance, cls).setUpClass()
        cls.test_object = TestObject.objects.create(name='TestName1')
        cls.test_object2 = TestObject.objects.create(name='TestName2')
        cls.deleted_obj = TestObject.objects.create(name='TestName3', state='deleted')
        cls.deleted_obj2 = TestObject.objects.create(name='TestName4', status='deleted')

    def test_convert_to_one_instance(self):
        class LoadSchema(Schema):
            test_object = fields.Instance(TestObject)

        result = LoadSchema().load({"test_object": str(self.test_object.id)})
        self.assertIn("test_object", result)
        self.assertEqual(result["test_object"].id, self.test_object.id)

    def test_convert_to_one_instance_allow_deleted(self):
        class LoadSchemaAllowDeleted(Schema):
            test_object = fields.Instance(TestObject, allow_deleted=True)

        result = LoadSchemaAllowDeleted().load({"test_object": str(self.deleted_obj.id)})
        self.assertIn("test_object", result)
        self.assertEqual(result["test_object"].id, self.deleted_obj.id)

    def test_convert_to_one_instance_check_deleted_by(self):
        class LoadSchemaCheckDeletedBy(Schema):
            test_object = fields.Instance(TestObject, allow_deleted=True, check_deleted_by='status')

        result = LoadSchemaCheckDeletedBy().load({"test_object": str(self.deleted_obj2.id)})
        self.assertIn("test_object", result)
        self.assertEqual(result["test_object"].id, self.deleted_obj2.id)

    def test_convert_to_one_instance_not_allow_deleted(self):
        class LoadSchemaNotAllowDeleted(Schema):
            test_object = fields.Instance(TestObject)
        try:
            LoadSchemaNotAllowDeleted().load({"test_object": str(self.deleted_obj.id)})
        except ValidationError as exc:
            self.assertIn("test_object", exc.messages)
            self.assertIn(f'Could not find document.', exc.messages["test_object"])

    def test_convert_to_one_instance_return_field(self):
        class LoadSchemaReturnField(Schema):
            test_object = fields.Instance(TestObject, return_field='id')

        result = LoadSchemaReturnField().load({"test_object": str(self.test_object.id)})
        self.assertIn("test_object", result)
        self.assertEqual(result["test_object"], self.test_object.id)

    def test_convert_to_one_instance_return_field_not_found_field(self):
        class LoadSchemaReturnFieldNotFound(Schema):
            test_object = fields.Instance(TestObject, return_field='not_found')
        try:
            LoadSchemaReturnFieldNotFound().load({"test_object": str(self.test_object.id)})
        except ValidationError as exc:
            self.assertIn("test_object", exc.messages)
            self.assertIn("Not found in model this field: 'not_found'.", exc.messages["test_object"])

    def test_convert_to_one_instance_search_instance_by_field(self):
        class LoadSchemaField(Schema):
            test_object = fields.Instance(TestObject, field='name')

        result = LoadSchemaField().load({"test_object": str(self.test_object.name)})
        self.assertIn("test_object", result)
        self.assertEqual(result["test_object"].id, self.test_object.id)

    def test_convert_to_one_instance_search_instance_by_field_not_found_field(self):
        class LoadSchemaField(Schema):
            test_object = fields.Instance(TestObject, field='names')
        try:
            LoadSchemaField().load({"test_object": str(self.test_object.name)})
        except ValidationError as exc:
            self.assertIn("test_object", exc.messages)
            self.assertIn("Not found in model this field: 'names'.", exc.messages["test_object"])

    def test_convert_to_one_instance_invalid_id(self):
        class LoadSchemaField(Schema):
            test_object = fields.Instance(TestObject)
        invalid_id = '1a'
        try:
            LoadSchemaField().load({"test_object": invalid_id})
        except ValidationError as exc:
            self.assertIn("test_object", exc.messages)
            self.assertIn(f"Invalid identifier: '{invalid_id}'.", exc.messages["test_object"])

    def test_convert_many_instances(self):
        class LoadSchema(Schema):
            test_objects = fields.Instance(TestObject, many=True)

        result = LoadSchema().load({"test_objects": [str(self.test_object.id), str(self.test_object2.id)]})
        self.assertIn("test_objects", result)
        list_ids = [obj.id for obj in result['test_objects']]
        self.assertIn(self.test_object.id, list_ids)
        self.assertIn(self.test_object2.id, list_ids)

    def test_convert_many_instances_string_in_value(self):
        class LoadSchemaStringInValue(Schema):
            test_objects = fields.Instance(TestObject, many=True)

        result = LoadSchemaStringInValue().load({"test_objects": f"{str(self.test_object.id)}, "
                                                                 f"{str(self.test_object2.id)}"})
        self.assertIn("test_objects", result)
        list_ids = [obj.id for obj in result['test_objects']]
        self.assertIn(self.test_object.id, list_ids)
        self.assertIn(self.test_object2.id, list_ids)

    def test_convert_many_instances_invalid_ids(self):
        invalid_ids = {'1', '2'}

        class LoadSchemaInvalidIds(Schema):
            test_objects = fields.Instance(TestObject, many=True)
        try:
            LoadSchemaInvalidIds().load({"test_objects": invalid_ids})
        except ValidationError as exc:
            self.assertIn("test_objects", exc.messages)
            self.assertIn(f"Invalid identifier: '{invalid_ids}'.", exc.messages["test_objects"])

    def test_convert_many_instances_not_found(self):
        not_found_ids = ['555555555555555555555555']

        class LoadSchemaNotFoundMany(Schema):
            test_objects = fields.Instance(TestObject, many=True)
        try:
            LoadSchemaNotFoundMany().load({"test_objects": not_found_ids})
        except ValidationError as exc:
            self.assertIn("test_objects", exc.messages)
            self.assertIn(f"Could not find document.", exc.messages["test_objects"])

    def test_convert_many_instances_not_all_where_found(self):
        not_found_ids = [str(self.test_object.id), "555555555555555555555555"]

        class LoadSchemaNotAllWhereFound(Schema):
            test_objects = fields.Instance(TestObject, many=True, assert_every=True)

        try:
            LoadSchemaNotAllWhereFound().load({"test_objects": not_found_ids})
        except ValidationError as exc:
            self.assertIn("test_objects", exc.messages)
            self.assertIn(f"Not all documents were found.", exc.messages["test_objects"])

    def test_convert_many_instances_assert_every_empty_list(self):
        not_found_ids = []

        class LoadSchemaAssertEveryEmptyList(Schema):
            test_objects = fields.Instance(TestObject, many=True, assert_every=True)

        result = LoadSchemaAssertEveryEmptyList().load({"test_objects": not_found_ids})
        self.assertIn('test_objects', result)
        self.assertEqual(result['test_objects'], [])


if __name__ == '__main__':
    main()



import typing

from marshmallow.fields import *
from mongoengine import ValidationError as MongoValidationError, Document, QuerySet, InvalidQueryError, ObjectIdField
from mongoengine.base import TopLevelDocumentMetaclass


__all__ = ["Field", "Raw", "Nested", "Mapping", "Dict", "List", "Tuple", "String", "UUID", "Number", "Integer",
           "Decimal", "Boolean", "Float", "DateTime", "NaiveDateTime", "AwareDateTime", "Time", "Date", "TimeDelta",
           "Url", "URL", "Email", "IP", "IPv4", "IPv6", "Method", "Function", "Str", "Bool", "Int", "Constant",
           "Pluck", "Instance"]


class Instance(Field):
    sql_db = False
    value = None

    #: Default error messages.
    default_error_messages = {
        "not_found_doc": "Could not find document.",
        "invalid_id": "Invalid identifier: '{value}'.",
        "not_found_field": "Not found in model this field: '{field_name}'.",
        "not_found_all_doc": "Not all documents were found.",
    }

    def __init__(self,
                 model: TopLevelDocumentMetaclass,
                 many: bool = False,
                 field: str = 'pk',
                 allow_deleted: bool = False,
                 check_deleted_by: str = 'state',
                 assert_every: bool = False,
                 return_field: str = None, **kwargs):
        """
        Initialisation class

        :param model: Model
        :param field: Found instances by this field.
        :param allow_deleted Allowed return deleted instances flag
        :param check_deleted_by Filed, by check deleted instances. (If allow_deleted=False)
        :param return_field: Return value field in this instance
        :param many: Many instances. True/False
        :param assert_every: True/False. Raise exception if not found one instances. (Only many=True)
        :param kwargs:
        """
        super().__init__(**kwargs)

        self.model = model
        self.many = many
        self.field = field
        self.allow_deleted = allow_deleted
        self.check_deleted_by = check_deleted_by
        self.assert_every = assert_every
        self.return_field = return_field
        self.sql_db = self.__check_sql_db()
        self.object_id = ObjectIdField == self.model.id.__class__

    @classmethod
    def __check_sql_db(cls) -> bool:
        # TODO Добавить поверку модели
        cls._query_func = cls._query_nosql
        return False

    def _serialize(self, value, attr, obj, **kwargs) -> typing.Optional[str]:
        """For Schema().dump() func"""
        return None

    def _deserialize(self, value, attr, data, **kwargs) -> typing.Union[Document, QuerySet]:
        """
        For Schema().load() func

        :param value: Value
        :param attr: Attribute name
        :param data:
        :param kwargs: Other params
        :return: One instance or list instances

        Example:
            data = {"attribute_name": 123456}
            ...
            class ClassName(Schema):
                attribute_name = fields.ToInstance(User)
            ...
            result = ClassName().load(data)     # {"attribute_name": UserInstance}
        """
        if isinstance(value, list) and not value:
            return value
        self.value = value
        try:
            result = self._convert_to_many() if self.many else self._query_func().first()
        except MongoValidationError as exc:
            raise self.make_error("invalid_id", value=value)
        except InvalidQueryError:
            raise self.make_error("not_found_field", field_name=self.field)
        else:
            if not result:
                raise self.make_error("not_found_doc")
            return self._get_value(result) if self.return_field else result

    def _convert_to_many(self) -> typing.List[Document]:
        """
        Convert to many instances

        :return: QuerySet
        """
        values = self.value
        if isinstance(values, str):
            values = [item.strip() for item in values.split(',')]
        if isinstance(values, list):
            self.value = list(set(values))
            query = self._query_func()
            if self.assert_every and query.count() != len(values):
                raise self.make_error("not_found_all_doc")
            return list(query)
        else:
            raise MongoValidationError

    def _query_sql(self, *args, **kwargs):
        pass

    def _query_nosql(self) -> QuerySet:
        """
        Query with mongoengine

        :return: QuerySet
        """
        # Generate filter data
        query_field = f"{self.field}__in" if self.many else self.field
        filter_data = {query_field: self.value}
        if not self.allow_deleted:
            filter_data.update({f'{self.check_deleted_by}__ne': 'deleted'})
        # Query
        return self.model.objects.filter(**filter_data)

    def _get_value(self, instance: typing.Union[Document, typing.List[Document]]):
        """
        Get value from founded instances

        :param instance: Convert result
        :return: value field in this instance
        """

        instances = instance if self.many else [instance]

        fields = self.model.columns if self.sql_db else getattr(self.model, "_fields").keys()
        if self.return_field in fields:
            result = [getattr(doc, self.return_field) for doc in instances]
            return result if self.many else result[0]
        raise self.make_error("not_found_field", field_name=self.return_field)



from mongoengine import Document, StringField


class TestObject(Document):

    name = StringField()
    state = StringField(choices=['active', 'deleted'], default='active')
    status = StringField(choices=['active', 'deleted'], default='active')

    meta = {
        "collection": "test_objects"
    }



from os import getenv
from unittest import TestCase
from dotenv import load_dotenv
from mongoengine import connect


load_dotenv()


class CommonTestCase(TestCase):
    db = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.db = connect(host=getenv("MONGODB_HOST"))
        if 'test' in cls.db.list_database_names():
            cls.db.drop_database('test')

    @classmethod
    def tearDownClass(cls) -> None:
        pass
